/*********************************************************************
 * 
 *    Filename:     cpad.c
 *    Author:       Ryan Shillington <shillington@writeme.com>
 *    Description:  Corel Palm Administrator Deamon
 *    Details:      This application serves requests back to clients
 *                  (Just palm pilots using IR, for now)
 *  
 *    This program is free software; you can redistribute it and/or
 *    modify it under the terms of the GNU General Public License as
 *    published by the Free Software Foundation; either version 2 of
 *    the License, or (at your option) any later version.
 *
 ********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <netinet/in.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <string.h>
#include <fcntl.h>
#include <errno.h>
#include <unistd.h>

#include "getOutput.h"
#include "network.h"
#include "cpad.h"

int irport;
char output[RX_BUF_SIZE];

int main(int argc, char **argv)
{
  char request[RX_BUF_SIZE];
  char data[TX_BUF_SIZE];
  char response[RX_BUF_SIZE];
  int result;
  int dest_id=0;
  int tries;

  if(argc != 1)
  {
    fprintf(stderr, "Usage: cpad\n");
    exit(1);
  }

  //* Open the irport for transfer
  if ((irport = open("/dev/irport", O_RDWR)) == -1)
  {
    DEBUG(0, "Error opening /dev/irport.\nErrno = %d\n", errno);
    exit(1);
  }
  
  //* Keep servicing requests
  while(1)
  {
    //* Initialize important variables 
    memset(request, 0x0, RX_BUF_SIZE);
    memset(response, 0x0, RX_BUF_SIZE);

    //* Get a request ('read' will block)
    DEBUG(1, "Searching for a request...\n");
    result = read(irport, request, RX_BUF_SIZE);
    request[result] = '\0';
  
    //* Get the destination ID
    dest_id = *((int *)(request+2));
    dest_id = ntohl(dest_id);
  
    DEBUG(4, "Received Command: %s\n", request+10);
    
    //* Send a response
    DEBUG(4, "Sending Response to command...\n");
    WRITE(irport, "%c%c%c%c%c%c%c%c%c%c%s", 
          0xFF, 0xF3, 
          (CPAD_ADDR & 0xFF000000) >> 24, 
          (CPAD_ADDR & 0x00FF0000) >> 16, 
          (CPAD_ADDR & 0x0000FF00) >> 8, 
          CPAD_ADDR & 0x000000FF, 
          (dest_id & 0xFF000000) >> 24, 
          (dest_id & 0x00FF0000) >> 16, 
          (dest_id & 0x0000FF00) >> 8, 
          dest_id & 0x000000FF, 
          request+10);

  
    //* Fetch the data that the client wants
    serviceRequest(request+10, data);

    //* Send out the data until we get a response
    tries = 0;
    while((strncmp(response+10, data, strlen(data)) != 0) && (tries < 6))
    {
      DEBUG(4, "Sending... Data = %s\n", data);
      WRITE(irport, "%c%c%c%c%c%c%c%c%c%c%s", 
          0xFF, 0xF3, 
          (CPAD_ADDR & 0xFF000000) >> 24, 
          (CPAD_ADDR & 0x00FF0000) >> 16, 
          (CPAD_ADDR & 0x0000FF00) >> 8, 
          CPAD_ADDR & 0x000000FF, 
          (dest_id & 0xFF000000) >> 24, 
          (dest_id & 0x00FF0000) >> 16, 
          (dest_id & 0x0000FF00) >> 8, 
          dest_id & 0x000000FF, 
          data);

      sleep(1);

      //* Get a response 
      fcntl(irport, F_SETFL, O_NONBLOCK);
      if((result = read(irport, response, RX_BUF_SIZE)) != 0)
      {
        DEBUG(4, "Received Response\n");
      } else {
        DEBUG(0, "Could not read IR information.\nErrno = %d\n", errno);
      }
      fcntl(irport, F_SETFL, 0);
      tries++;
    }
    
    DEBUG(4, "irport = %d, CPAD_ADDR = %x, dest_id = %x, errno=%d\n", 
        irport, CPAD_ADDR, dest_id, errno);
    
  }

  close(irport);
  return 0;
}

void serviceRequest(char request[RX_BUF_SIZE], char data[TX_BUF_SIZE])
{
  //* Figure out what the request is:
  switch(request[0]) {
    case 'C': 
      //* Request is for Changing IP Address Information
      serviceChangeIP(request[12], data);
      break;
    case 'E':
      //* Request is to EXECUTE a Command
      serviceExecute(&request[8], data);
      break;
    case 'L': //* Request is for Load average
      serviceLoad(data);
      break;
    case 'M': //* Request is for Memory info
      serviceMemory(data);
      break;
    case 'S': //* Request is to SET attributes on the netwinder
      switch(request[1]) {
        case 'C': //* Request is to SET IP info
          serviceSetIP(request);
          break;
      }
      data[0] = '\0';
      break;
    case 'T': //* Request is for Temperature info
      serviceTemperature(data);
      break;
    default:
      DEBUG(0, "Unknown request received. Request = %s\n", request);
  }
}

void serviceExecute(char *request, char data[TX_BUF_SIZE])
{
  char *args[64];     //* 64 arguments should be plenty
  char *filename;
  int find = 0;
  int curr = 0;
  char *findp;

  //* The filename should be the first item
  filename = args[curr++] = request;
  findp = strchr(request, ' ');
  
  //* Find any and all arguements
  while(findp != NULL)
  {
    findp[0] = '\0';
    args[curr++] = findp+1;
    findp = strchr(findp+1, ' ');
  }
  args[curr++] = NULL;
  
  //* Find the carriage return and mark it as the end of the command
  find = 0;
  while(request[find] != '\n') find++;
  request[find] = '\0';

  //* Execute the command and get the output
  getOutput(filename, args, data);
  find = strlen(data);
  data[find] = '\0';
}

void serviceSetIP(char request[RX_BUF_SIZE])
{
  char device[20], hostname[64], alias[64];
  byte IP[4], netmask[4], gateway[4];
  int find = 0;
  int curr = 0;
  
  //* Get the DEVICE
  while(request[find] != '=') find++;
  find++;
  while(request[find] != ',') device[curr++] = request[find++];
  device[curr] = '\0';
  find++;

  //* Get the IP ADDRESS
  readIP(request+find, IP);
  while(request[find] != ',') find++;
  find++;

  //* Get the FULL NAME (if there is one)
  curr=0;
  while(request[find] != ',') hostname[curr++] = request[find++];
  hostname[curr] = '\0';
  find++;

  //* Get the ALIAS (if there is one)
  curr=0;
  while(request[find] != ',') alias[curr++] = request[find++];
  alias[curr] = '\0';
  find++;

  //* Get the NETMASK
  readIP(request+find, netmask);
  while(request[find] != ',') find++;
  find++;

  //* Get the GATEWAY
  readIP(request+find, gateway);
  while(request[find] != ',') find++;
  find++;
  
  //* Do a sanity check to make sure we haven't run over the end of the
  // buffer.
  curr = (strchr(request, '\n') - request);
  DEBUG(4, "device = %s, hostname = %s, alias = %s, find = %d, curr=%d\n", 
      device, hostname, alias, find, curr);
  if(find != curr)
  {
    fprintf(stderr, "ERROR: Invalid IP information was received.
        Discarding it...\n");
    return;
  }

  //* Set the IP information accordingly
  writeHostsFile(IP, hostname, alias);
  writeNetworkFile(hostname, gateway, device);
  writeIfcfgFile(netmask, IP, device);
  
  //* Restart the Network
}

//* A helper function to read IP addresses out of a char stream
void readIP(char *data, byte ip[4])
{
  int ipNum, find = 0;

  //* 1 - Find all 4 ip numbers
  for (ipNum = 0; ipNum < 4; ipNum++)
  {
    //* 1.1 - Convert the text to a number
    ip[ipNum] = atoi(data+find);

    //* 1.2 - Find the next number
    while(data[find] != '.') find++;
    find++;
  }
}

void serviceTemperature(char data[TX_BUF_SIZE])
{
  int therm;                      //* File discriptor
  char input[160];
  int find, curr;
  
  //* 1 - Open the /proc/therm to fetch data
  if ((therm = open("/proc/therm", O_RDONLY)) == -1)
  {
    DEBUG(0, "Error opening /proc/therm.\nErrno = %d\n", errno);
    exit(1);
  }
  
  //* 2 - Get the data
  read(therm, input, 160);
  DEBUG(4, "Therm = %s", input);
  
  //* 3 - Transfer the HI temperature
  find = 0;
  curr = 0;
  //* 3.1 - Figure out where the data starts
  while (input[find++] != 'I');
  find++;
  //* 3.2 - Figure out where the data ends (and copy it)
  while (input[find] != ',') 
    data[curr++] = input[find++];
  data[curr++] = ',';
  
  //* 4 - Transfer the LO temperature
  //* 4.1 - Figure out where the data starts
  while (input[find++] != 'O');
  find++;
  //* 4.2 - Figure out where the data ends (and copy it)
  while (input[find] != ';') 
    data[curr++] = input[find++];
  data[curr++] = ',';
  
  //* 5 - Transfer the current temperature
  //* 5.1 - Figure out where the data starts
  while (input[find++] != ':');
  find++;
  //* 5.2 - Figure out where the data ends (and copy it)
  while (input[find] != ' ') 
    data[curr++] = input[find++];
  data[curr++] = ',';

  //* 6 - Transfer the fan setting
  //* 6.1 - Figure out where the data starts
  while (input[find++] != 'n');
  find++;
  //* 6.2 - Figure out where the data ends (and copy it)
  while (input[find] != '.') 
    data[curr++] = input[find++];
  data[curr++] = '\0';
  
  close(therm);
}

void serviceMemory(char data[TX_BUF_SIZE])
{
  int memory;                      //* File discriptor
  char input[512];
  int i, find, curr;

  //* 1 - Open the /proc/memory to fetch data
  if ((memory = open("/proc/meminfo", O_RDONLY)) == -1)
  {
    DEBUG(0, "Error opening /proc/memory.\nErrno = %d\n", errno);
    exit(1);
  }
  
  //* 2 - Get the data
  read(memory, input, 512);
  DEBUG(4, "Therm = %s", input);
  
  //* 3 - Find the TOTAL MEMORY
  find = 0;
  curr = 0;
  //* 3.1 - Figure out where the data starts
  for(i = 0; i < 7; i++, find++)
    while (input[find] != ':') find++;
  while (input[find] == ' ') find++;
  //* 3.2 - Figure out where the data ends (and copy it)
  while (input[find] != ' ') 
    data[curr++] = input[find++];
  data[curr++] = ',';

  //* 4 - Find the USED MEMORY
  //* 4.1 - Figure out where the data starts
  while (input[find] == ' ') find++;
  //* 4.2 - Figure out where the data ends (and copy it)
  while (input[find] != ' ') 
    data[curr++] = input[find++];
  data[curr++] = ',';
  
  //* 5 - Find the FREE MEMORY
  //* 5.1 - Figure out where the data starts
  while (input[find] == ' ') find++;
  //* 5.2 - Figure out where the data ends (and copy it)
  while (input[find] != ' ') 
    data[curr++] = input[find++];
  data[curr++] = ',';

  //* 6 - Find the SHARED MEMORY
  //* 6.1 - Figure out where the data starts
  while (input[find] == ' ') find++;
  //* 6.2 - Figure out where the data ends (and copy it)
  while (input[find] != ' ') 
    data[curr++] = input[find++];
  data[curr++] = ',';

  //* 7 - Find the BUFFERED MEMORY
  //* 7.1 - Figure out where the data starts
  while (input[find] == ' ') find++;
  //* 7.2 - Figure out where the data ends (and copy it)
  while (input[find] != ' ') 
    data[curr++] = input[find++];
  data[curr++] = ',';

  //* 8 - Find the CACHED MEMORY
  //* 8.1 - Figure out where the data starts
  while (input[find] == ' ') find++;
  //* 8.2 - Figure out where the data ends (and copy it)
  while ((input[find] != '\n') && (input[find] != ' '))
    data[curr++] = input[find++];
  data[curr++] = ',';

  //* 9 - Find the TOTAL SWAPPED MEMORY
  //* 9.1 - Figure out where the data starts
  while (input[find] != ':') find++;
  find++;
  while (input[find] == ' ') find++;
  //* 9.2 - Figure out where the data ends (and copy it)
  while (input[find] != ' ') 
    data[curr++] = input[find++];
  data[curr++] = ',';

  //* 10 - Find the USED SWAPPED MEMORY
  //* 10.1 - Figure out where the data starts
  while (input[find] == ' ') find++;
  //* 10.2 - Figure out where the data ends (and copy it)
  while (input[find] != ' ') 
    data[curr++] = input[find++];
  data[curr++] = ',';

  //* 11 - Find the FREE SWAPPED MEMORY
  //* 11.1 - Figure out where the data starts
  while (input[find] == ' ') find++;
  //* 11.2 - Figure out where the data ends (and copy it)
  while ((input[find] != ' ') && (input[find] != '\n'))
    data[curr++] = input[find++];
  data[curr++] = ',';
  data[curr++] = '\0';
  
  close(memory);
}

void serviceLoad(char data[TX_BUF_SIZE])
{
  char input[256];
  int find, curr;
  char *const args[] = { "uptime", NULL };
  
  //* 1 - Call uptime to get the data
  getOutput("uptime", args , input);
  
  //* 2 - Transfer the Time
  find = 0;
  curr = 0;
  //* 2.1 - Figure out where the data starts
  while (input[find] == ' ') find++;
  //* 2.2 - Figure out where the data ends (and copy it)
  while (input[find] != ' ') 
    data[curr++] = input[find++];
  data[curr++] = ',';
  
  //* 3 - Transfer the UPTIME
  //* 3.1 - Figure out where the data starts (skip the 'up')
  while (input[find] == ' ') find++;
  while (input[find] != ' ') find++;
  while (input[find] == ' ') find++;
  //* 3.2 - Figure out where the data ends (and copy it)
  while (input[find] != ',') 
    data[curr++] = input[find++];
  find++;
  data[curr++] = ',';
  
  //* 4 - Transfer the NUMBER OF USERS
  //* 4.1 - Figure out where the data starts (This has to be funky because the
  //layout of the uptime always changes).
  while (input[find] != 'u') find++;
  find--; find--;
  while (input[find] != ' ') find--;
  find++;
  //* 4.2 - Figure out where the data ends (and copy it)
  while (input[find] != ' ') 
    data[curr++] = input[find++];
  data[curr++] = ',';

  //* 5 - Transfer the 1 MIN LOAD AVG
  //* 5.1 - Figure out where the data starts
  while (input[find] != ':') find++;
  find++;
  while (input[find] == ' ') find++;
  //* 5.2 - Figure out where the data ends (and copy it)
  while (input[find] != ',') 
    data[curr++] = input[find++];
  data[curr++] = ',';
  find++;

  //* 6 - Transfer the 5 MIN LOAD AVG
  //* 6.1 - Figure out where the data starts
  while (input[find] == ' ') find++;
  //* 6.2 - Figure out where the data ends (and copy it)
  while (input[find] != ',') 
    data[curr++] = input[find++];
  data[curr++] = ',';
  find++;

  //* 7 - Transfer the 15 MIN LOAD AVG
  //* 7.1 - Figure out where the data starts
  while (input[find] == ' ') find++;
  //* 7.2 - Figure out where the data ends (and copy it)
  while ((input[find] != ' ') && (input[find] != '\n'))
    data[curr++] = input[find++];
  data[curr++] = ',';

  data[curr++] = '\0';
  
}

void serviceChangeIP(char id, char data[TX_BUF_SIZE])
{
  int hosts;                        //* File discriptor
  int network;                      //* File discriptor
  int ifcfg;                        //* File discriptor
  char hosts_input[4096], network_input[4096], ifcfg_input[4096]; //* Buffers
  int hosts_size, network_size, ifcfg_size;
  int find, curr;
  char *findp, *find2p;
  char *ifcfg_path;

  //* 1 - Open the files to read the data out
  if ((hosts = open(HOSTS_PATH, O_RDONLY)) == -1)
  {
    DEBUG(0, "Error opening %s.\nErrno = %d\n", HOSTS_PATH, errno);
    exit(1);
  }
  if ((network = open(NETWORK_PATH, O_RDONLY)) == -1)
  {
    DEBUG(0, "Error opening %s.\nErrno = %d\n", NETWORK_PATH, errno);
    exit(1);
  }
  
  //* 2 - Get the data
  hosts_size = read(hosts, hosts_input, 4096);
  hosts_input[hosts_size] = '\0';
  DEBUG(4, "hosts = %s\n", hosts_input);
  network_size = read(network, network_input, 4096);
  network_input[network_size] = '\0';
  DEBUG(4, "network = %s\n", network_input);
  
  //* 3 - Find the Device, depending on the ID given
  switch (id)
  {
    case '?':
      //* 3 - Find the DEVICE
      find = 0;
      curr = 0;
      //* 3.1 - Figure out where the data starts
      findp = strstr(network_input, "GATEWAYDEV");
      while (findp[find] != '=') find++;
      find++;
      //* 3.2 - Figure out where the data ends (and copy it)
      while ((findp[find] != ' ') && (findp[find] != '\n'))
        data[curr++] = findp[find++];
      data[curr++] = ',';

      //* 3.3 - figure out which configureation to open
      if(strncmp(data, "eth0", 4) == 0)
      {
        ifcfg_path = IFCFG_ETH0_PATH;
      } else if (strncmp(data, "eth1", 4) == 0) {
        ifcfg_path = IFCFG_ETH1_PATH;
      } else {
        data[4] = '\0';
        fprintf(stderr, "ERROR: Configuration %s is not understood\n", data);
        return;
      }
      break;
    case '0':
      //* 3 - Use ETH0
      ifcfg_path = IFCFG_ETH0_PATH;
      sprintf(data, "eth0,");
      curr = strlen(data);
      break;
    case '1':
      //* 3 - Use ETH1
      ifcfg_path = IFCFG_ETH1_PATH;
      sprintf(data, "eth1,");
      curr = strlen(data);
      break;
    default:
      fprintf(stderr, "ERROR: received invalid data\n");
      return;
  }
  
  //* 4 - Open the file... whichever one it is
  if ((ifcfg = open(ifcfg_path, O_RDONLY)) == -1)
  {
    DEBUG(0, "Error opening %s.\nErrno = %d\n", ifcfg_path, errno);
    exit(1);
  }
  ifcfg_size = read(ifcfg, ifcfg_input, 4096);
  ifcfg_input[ifcfg_size] = '\0';
  DEBUG(4, "ifcfg = %s\n", ifcfg_input);

  //* 5 - Find the IP ADDRESS of the host
  find = 0;
  //* 5.1 - Figure out where the data starts
  findp = strstr(ifcfg_input, "IPADDR");
  while(findp[find] != '=') findp++;
  findp++;
  //* 5.2 - Figure out where the data ends (and copy it)
  while ((findp[find] != ' ') && (findp[find] != '\n'))
    data[curr++] = findp[find++];
  data[curr++] = ',';

  //* 6 - Find the NAME
  //* 6.1 - Figure out where the data starts
  findp[find] = '\0';
  find2p = strstr(hosts_input, findp);
  //* Make sure that an entry actually exists
  if(find2p != NULL)
  {
    findp[find] = '\n'; find = (int) (find2p - hosts_input);
    while ((hosts_input[find] != ' ') && (hosts_input[find] != '\t')) find++;
    while ((hosts_input[find] == ' ') || (hosts_input[find] == '\t')) find++;
    //* 6.2 - Figure out where the data ends (and copy it)
    while (hosts_input[find] != ' ') 
      data[curr++] = hosts_input[find++];
    data[curr++] = ',';

    //* 7 - Find the ALIAS
    //* 7.1 - Figure out where the data starts
    while ((hosts_input[find] == ' ') || (hosts_input[find] == '\t')) find++;
    //* 7.2 - Figure out where the data ends. Only copy if there is an alias
    //before the end of the line
    if(hosts_input[find] != '\n')
      while ((hosts_input[find] != ' ') && (hosts_input[find] != '\n'))
      data[curr++] = hosts_input[find++];
    data[curr++] = ',';
  } else {
    data[curr++] = ',';
    data[curr++] = ',';
  }

  //* 8 - Find the NETMASK
  find = 0;
  //* 8.1 - Figure out where the data starts
  findp = strstr(ifcfg_input, "NETMASK");
  while(findp[find] != '=') find++;
  find++;
  //* 8.2 - Figure out where the data ends (and copy it)
  while ((findp[find] != ' ') && (findp[find] != '\n'))
    data[curr++] = findp[find++];
  data[curr++] = ',';

  //* 8 - Find the GATEWAY
  find = 0;
  //* 8.1 - Figure out where the data starts
  findp = strstr(network_input, "GATEWAY");
  while(findp[find] != '=') find++;
  find++;
  //* 8.2 - Figure out where the data ends (and copy it)
  while ((findp[find] != ' ') && (findp[find] != '\n'))
    data[curr++] = findp[find++];
  data[curr++] = ',';
  
  data[curr] = '\0';
}

